
#include "ewrte.h"
#include "ewgfx.h"

/* Include the QR-code generator library */
#include "qrcodegen.h"


EW_MODULE
( 
  INTRINSICS_IFC_VERSION,
  L"IntrinsicsModule", 
  L"Extern Bitmap intrinsics module"
)


/* The function creates a new ALPHA8 bitmap with a QR-code generated from the
   given text aName. */
XBitmap* EwLoadExternBitmap( XString aName )
{
  static uint8_t qr0[ qrcodegen_BUFFER_LEN_MAX ];
  static uint8_t tempBuffer[ qrcodegen_BUFFER_LEN_MAX ];
  static char    text[ 4296 ];
  int            len = EwGetStringLength( aName );
  int            i;
  XBitmap*       bitmap;
  XPoint         frameSize;
  XRect          lockArea;
  XBitmapLock*   lock;
  unsigned char* dest;
  int            ofs;
  int            x, y;
  int            scale = 1;

  /* No text to encode or the text is too long? */
  if ( !len || ( len > ( sizeof( text ) - 1 )))
    return 0;

  /* Convert the wide-char text in ASCII text. This is not the optimal solution
     because it works for ASCII characters only. */
  for ( i = 0; i < len; i++ )
    if ( aName[i] < 128 ) text[i] = (char)aName[i];
    else                  text[i] = '?';

  /* Terminate the ASCII string. */
  text[i] = 0;

  /* Generate the QR-code. */
  if ( !qrcodegen_encodeText( text, tempBuffer, qr0, qrcodegen_Ecc_MEDIUM, 
        qrcodegen_VERSION_MIN, qrcodegen_VERSION_MAX, qrcodegen_Mask_AUTO, 1 ))
    return 0;

  /* Determine a scaling factor for the resulting image to fill the entire 360
     pixel as good as possible. */
  scale = 360 / qrcodegen_getSize( qr0 );

  if ( scale < 1 )
    scale = 1;

  /* Get the size of the QR-image in pixel */
  frameSize.X       = qrcodegen_getSize( qr0 ) * scale;
  frameSize.Y       = qrcodegen_getSize( qr0 ) * scale;
  lockArea.Point1.X = 0;
  lockArea.Point1.Y = 0;
  lockArea.Point2.X = frameSize.X;
  lockArea.Point2.Y = frameSize.Y;

  /* Create an Embedded Wizard bitmap to store the QR image pixel data */
  bitmap = EwCreateBitmap( EW_PIXEL_FORMAT_ALPHA8, frameSize, 0, 1 );

  /* Could not create the image. */
  if ( !bitmap )
    return 0;

  /* Obtain access to the bitmap. */
  lock = EwLockBitmap( bitmap, 0, lockArea, 0, 1 );
  dest = (unsigned char*)lock->Pixel1;
  ofs  = lock->Pitch1Y - frameSize.X;

  /* Fill the bitmap with QR image content. */
  for ( y = 0; y < frameSize.Y; y++, dest += ofs )
    for ( x = 0; x < frameSize.X; x++, dest++ )
      *dest = ( qrcodegen_getModule( qr0, x / scale, y / scale ))?
              0xFF : 0x00;

  /* Release the lock */
  EwUnlockBitmap( lock );

  return bitmap;
}


/* Table describing all intrinsics implemented in this module */
EW_DEFINE_INTRINSICS
  EW_INTRINSIC
  (
    L"EwLoadExternBitmap",
    L"handle",
    1,
    L"string",
    L"aName",
    EwLoadExternBitmap
  )
EW_END_OF_INTRINSICS


/* In order to create and load bitmaps the intrinsics module will need an access
   to the Graphics Engine API. This API is implemented in a separate intrinsics
   module. To use it, pointers to the API functions will be obtained and stored. */
static void* EwCreateBitmapProc = 0;
static void* EwLockBitmapProc   = 0;
static void* EwUnlockBitmapProc = 0;


/* Wrapper to the Graphics Engine API function EwCreateBitmap(). This function is
   implemented in the separate 'GraphicsEngine' intrinsics module. Its job is to
   create a new empty bitmap. */
XBitmap* EwCreateBitmap( int aFormat, XPoint aFrameSize, XInt32 aFrameDelay,
  XInt32 aNoOfFrames )
{
  typedef XBitmap* (*tmp)( int aFormat, XPoint aFrameSize, XInt32 aFrameDelay,
                           XInt32 aNoOfFrames );

  /* Only, if the entry point to the API function is known */
  if ( EwCreateBitmapProc )
    return ((tmp)EwCreateBitmapProc)( aFormat, aFrameSize, aFrameDelay, aNoOfFrames );

  return 0;
}


/* Wrapper to the Graphics Engine API function EwLockBitmap(). This function is
   implemented in the separate 'GraphicsEngine' intrinsics module. Its job is to
   lock the bitmap for direct memory access. */
XBitmapLock* EwLockBitmap( XBitmap* aBitmap, XInt32 aFrameNo, XRect aArea,
  XBool aRead, XBool aWrite )
{
  typedef XBitmapLock* (*tmp)( XBitmap* aBitmap, XInt32 aFrameNo, XRect aArea,
                               XBool aRead, XBool aWrite );

  /* Only, if the entry point to the API function is known */
  if ( EwLockBitmapProc )
    return ((tmp)EwLockBitmapProc)( aBitmap, aFrameNo, aArea, aRead, aWrite );

  return 0;
}


/* Wrapper to the Graphics Engine API function EwUnlockBitmap(). This function is
   implemented in the separate 'GraphicsEngine' intrinsics module. Its job is to
   unlock the bitmap after the direct memory access is done. */
void EwUnlockBitmap( XBitmapLock* aLock )
{
  typedef void (*tmp)( XBitmapLock* aLock );

  /* Only, if the entry point to the API function is known */
  if ( EwUnlockBitmapProc )
    ((tmp)EwUnlockBitmapProc)( aLock );
}


/* The function is called after all intrinsics modules has finalized their
   initialization. It determines the entry points for few intrinsics found
   in the foreign module 'GraphicsEngine' and needed to create bitmaps. */
EW_ENTRYPOINT void EwLinkModule( XModuleContext* aList, void* aContext )
{
  /* First search all loaded intrinsics modules for the 'GraphicsEngine'. */
  XModuleContext* module = EwFindModule( aList, "GraphicsEngine" );

  /* Then query the desired entry points and store them within global variables. */
  if ( module )
  {
    EwCreateBitmapProc = EwFindProc( module, "EwCreateBitmap" );
    EwLockBitmapProc   = EwFindProc( module, "EwLockBitmap"   );
    EwUnlockBitmapProc = EwFindProc( module, "EwUnlockBitmap" );
  }  
}


/* pba */